/*LINTLIBRARY */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <math32.h>
#include <complex.h>

#ifdef SKYBOLT
#define M_PI    _PI
#endif

#define TWO_PI    (2.0*M_PI)
typedef complex_float complex;
static complex *co_table;
static complex *exp_table;
static int nold = 0;

int 
icycle(i,nshift,length)
int i,nshift,length;
{
  while(1) {
    if (i >= nshift) break;
    i = i+length;
  }
  return(i-nshift);
}

int 
ilog2(n)
int n;
{
  int i;

  i=-1;
  while (1) {
    if (n <= 0) break;
    n=n>>1;
    i++;
  }
  return (i);
}

static int
bitreversal(k,n)
int k;
int n;
{
  int itemp,j,i;
  
  itemp=0;
  j=k;
  for (i=0; i < n; i++) {
    itemp=(itemp << 1) + (j & 1);
    j=j >> 1;
  }
  return (itemp);
}
  
static float Parzen(n,j)
int n,j;
{
   float fret,ftemp;

   ftemp=0.5*(n-1.0);
   fret = 1.0 - fabs((j -ftemp)/ftemp);
   return(fret);
}

static float Hanning(n,j)
int n,j;
{
   float fret;

   fret = 0.5 * (1.0 - rcos(TWO_PI * j / (n-1.0)));
   return(fret);
}

static float Hamming(n,j)
int n,j;
{
   float fret;

   fret = 0.54 - 0.46 * rcos(TWO_PI * j / (n-1.0));
   return(fret);
}

static float ExactBlackman(n,j)
int n,j;
{
   float fret,ftemp;

   ftemp=TWO_PI*j/(n-1.0);
   fret = 0.42 - 0.50 * rcos(ftemp) + 0.08 * rcos(2.0*ftemp);
   return(fret);
}

static float Welch(n,j)
int n,j;
{
   float fret;

   fret = (j - 0.5 * (n - 1.0)) / (0.5 * (n + 1.0));
   fret = 1.0 - fret*fret;
   return(fret);
}

int
fcvectFFTwindowtable(window,numdat,w)
int window,numdat;
float *w;
{
  int i;
  float multiplier;
  
  switch ( window ) {
  case 0:
  default:
    for ( i = 0; i < numdat ; ++i ) 
      w[i]=1.0;
    break;
  case 1:
    for ( i = 0; i < numdat ; ++i )
      w[i]=Parzen(numdat,i);
    break;
  case 2:
    for ( i = 0; i < numdat ; ++i )
      w[i]=Hanning(numdat,i);
    break;
  case 3:
    for ( i = 0; i < numdat ; ++i ) 
      w[i]=Welch(numdat,i);
    break;
  case 4:
    for ( i = 0; i < numdat ; ++i ) 
      w[i]=Hamming(numdat,i);
    break;
  case 5:
    for ( i = 0; i < numdat ; ++i ) 
      w[i]=ExactBlackman(numdat,i);
    break;
  }
}


int
fcvectFFTwindow(x,numdat,window,y)
complex *x,*y;
int numdat,window;
{
   int i;
   float multiplier;

   for ( i = 0; i < numdat ; ++i ) {
      switch ( window ) {
         case 0:
            multiplier = 1.0;
            break;
         case 1:
            multiplier = Parzen(numdat,i);
            break;
         case 2:
            multiplier = Hanning(numdat,i);
            break;
         case 3:
            multiplier = Welch(numdat,i);
            break;
         case 4:
            multiplier = Hamming(numdat,i);
            break;
         case 5:
            multiplier = ExactBlackman(numdat,i);
            break;
         default:
            multiplier = 1.0;
            break;
      }
      y[i] =fcmplx2op(fcmplx(multiplier,0.0),x[i],'*');
   }
}

int 
fcvectFFTtable(numdat,table)
int numdat;
complex *table;
{
  fcvectTable(numdat,table);
  nold=numdat;
  co_table=(complex *)table;
  return (0);
}

int 
fcvectEXPtable(numdat,table)
int numdat;
complex *table;
{
  fcvectTable(numdat,table);
  exp_table=(complex *)table;
  return (0);
}

int 
fcvectTable(numdat,table)
int numdat;
complex *table;
{
  float delta,theta,ftemp1,ftemp2;
  int i,n,n2,n3;

  delta=TWO_PI/numdat;
  theta=0.0;
  if (numdat % 4 == 0) {
    n=numdat/4;
    n2=2*n;
    n3=3*n;
    for (i=0; i < n; i++) {
      ftemp1=rsin(theta);
      ftemp2=rcos(theta);
      table[i]=fcmplx(ftemp2,ftemp1);
      table[i+n]=fcmplx(-ftemp1,ftemp2);
      table[i+n2]=fcmplx(-ftemp2,-ftemp1);
      table[i+n3]=fcmplx(ftemp1,-ftemp2);
      theta += delta;
   }
  }
  else {
    if (numdat % 2 == 0) {
      n=numdat/2;
      for (i=0; i < n; i++) {
	ftemp1=rsin(theta);
	ftemp2=rcos(theta);
	table[i]=fcmplx(ftemp2,ftemp1);
	table[i+n]=fcmplx(-ftemp2,-ftemp1);
	theta += delta;
      }
    }
    else {
      n=numdat;
      for (i=0; i < n; i++) {
	ftemp1=rsin(theta);
	ftemp2=rcos(theta);
	table[i]=fcmplx(ftemp2,ftemp1);
	theta += delta;
      }
    }
  }

  return(0);
}

int
fcvectFFTforward(x,numdat,y)
complex *x, *y;
int numdat;
{
  int i,j,n;
  complex ctemp;

#ifdef SKYBOLT
  if (numdat < 1025)
    v$_cvffts1(x,1,y,1,numdat,1);
  else
    v$_cvfftb(x,y,numdat,1);
#else
  for (j=0; j < numdat; j++) {
    ctemp=fcmplx(0.0,0.0);
    for (i=0; i < numdat; i++) {
      n=(i*j)%numdat;
      ctemp=fcmplx2op(ctemp,fcmplx2op(x[i],co_table[n],'@'),'+');
    }
    y[j]=ctemp;
  }
#endif
  return (0);
}

int
fcvectFFTforward2(x,numdat,y)
complex *x, *y;
int numdat;
{
#ifdef SKYBOLT
  if (numdat < 1024)
    v$_cvffts1(x,1,y,1,numdat,1);
  else
    v$_cvfftb(x,y,numdat,1);
#else
  if ((nold == 0) || (nold != numdat)) {
    if (co_table != NULL) free((char *)co_table);
    co_table=(complex *)malloc((unsigned)(numdat*sizeof(complex)));
    fcvectFFTtable(numdat,co_table);
   }
  memcpy(y,x,numdat*sizeof(complex));
  fft(numdat,y);
#endif
  
  return(0);
}

int
fcvectFFTforward0(x,numdat,y)
complex *x, *y;
int numdat;
{
  int i,j,n;
  complex ctemp;

  for (j=0; j < numdat; j++) {
    ctemp=fcmplx(0.0,0.0);
    for (i=0; i < numdat; i++) {
      n=(i*j)%numdat;
      ctemp=fcmplx2op(ctemp,fcmplx2op(x[i],co_table[n],'@'),'+');
    }
    y[j]=ctemp;
  }
  return (0);
}

int
fcvectFFTinverse(x,numdat,y)
complex *x, *y;
int numdat;
{
  int i,j,n;
  complex ctemp;

#ifdef SKYBOLT
  if (numdat < 1024)
    v$_cvffts1(x,1,y,1,numdat,-1);
  else
    v$_cvfftb(x,y,numdat,-1);
#else
  for (j=0; j < numdat; j++) {
    ctemp=fcmplx(0.0,0.0);
    for (i=0; i < numdat; i++) {
      n=(i*j)%numdat;
      ctemp=fcmplx2op(ctemp,fcmplx2op(x[i],co_table[n ],'*'),'+');
    }
    y[j]=fcmplx2op(ctemp,fcmplx(1.0/numdat,0.0),'*');
  }
#endif
  return (0);
}

int
fcvectFFTinverse2(x,numdat,y)
complex *x, *y;
int numdat;
{
#ifdef SKYBOLT
  if (numdat < 1024)
    v$_cvffts1(x,1,y,1,numdat,-1);
  else
    v$_cvfftb(x,y,numdat,-1);
#else
  if ((nold == 0) || (nold != numdat)) {
    if (co_table != NULL) free((char *)co_table);
    co_table=(complex *)malloc((unsigned)(numdat*sizeof(complex)));
    fcvectFFTtable(numdat,co_table);
   }
  memcpy(y,x,numdat*sizeof(complex));
  ifft(numdat,y);
#endif
  return (0);
}

int
fcvectFFTinverse0(x,numdat,y)
complex *x, *y;
int numdat;
{
  int i,j,n;
  complex ctemp;

  for (j=0; j < numdat; j++) {
    ctemp=fcmplx(0.0,0.0);
    for (i=0; i < numdat; i++) {
      n=(i*j)%numdat;
      ctemp=fcmplx2op(ctemp,fcmplx2op(x[i],co_table[n ],'*'),'+');
    }
    y[j]=fcmplx2op(ctemp,fcmplx(1.0/numdat,0.0),'*');
  }
  return (0);
}

/* solve a matrix equation by Gaussian Elimination */
int
fmateqs(a,y,n,eps,x)
float *a,*y,*x;
int n;
float eps;
{
  int i,j,k,itemp,itemp1,itemp2,n_1,n1,imax;
  float ftemp,max,*am;

  n_1=n-1;
  n1=n+1;
  am=(float *)malloc((unsigned)n*n1*sizeof(float));
  if (am == NULL) {
    fprintf(stderr, "Cannot allocate memory in fdsp for fmateqs.\n");
    return (2);
  }
  
  memcpy((char *)am,(char *)a,n*n*sizeof(float));
  memcpy((char *)&am[n*n],(char *)y,n*sizeof(float));

  for (k=0; k < n_1; k++) {
    itemp=k*n;
    max=fabs(am[k+itemp]);
    imax=k;
    for (i=k; i < n; i++) {
      ftemp=fabs(am[i+itemp]);
      if (ftemp > max) {
	max=ftemp;
	imax=i;
      }
    }
    if (max < eps) return (1);
    max=am[imax+itemp];
    if (imax != k) {
      for (j=k; j < n1; j++) {
	itemp1=k+j*n;
	itemp2=imax+j*n;
	ftemp=am[itemp1];
	am[itemp1]=am[itemp2];
	am[itemp2]=ftemp;
      }
    }
    for (j=k; j < n1; j++) {
      itemp1=k+j*n;
      am[itemp1]=am[itemp1]/max;
    }

    for (i=k+1; i < n; i++) {
      for (j=k+1; j < n1; j++) {
	itemp1=i+j*n;
	am[itemp1] = am[itemp1]-am[k+j*n]*am[i+itemp];
      }
    }
  }
  
  itemp=n*n;
  x[n_1]=am[itemp+n_1]/am[itemp-1];
  for (k=n_1-1; k >= 0; k--) {
    ftemp=am[k+itemp];
    for (j=k+1; j < n; j++) 
      ftemp -= x[j]*am[k+j*n];
    x[k]=ftemp;
  }
  
  free((char *)am);
  return(0);
}
      
/* solve a complex matrix eqaution by Gaussian Elimination */
int
fcmateqs(a,y,n,eps,x)
complex *a,*y,*x;
int n;
float eps;
{
  int i,j,k,itemp,itemp1,itemp2,n_1,n1,imax;
  float ftemp,max;
  complex *am,cmax,ctemp;

  n_1=n-1;
  n1=n+1;
  am=(complex *)malloc((unsigned)n*n1*sizeof(complex));
  if (am == NULL) {
    fprintf(stderr, "Cannot allocate memory in fdsp for fmateqs.\n");
    return (2);
  }
  
  memcpy((char *)am,(char *)a,n*n*sizeof(complex));
  memcpy((char *)&am[n*n],(char *)y,n*sizeof(complex));

  for (k=0; k < n_1; k++) {
    itemp=k*n;
    itemp1=k+itemp;
    max=fabs(freal(am[itemp1]))+fabs(fimag(am[itemp1]));
    imax=k;
    for (i=k; i < n; i++) {
      itemp1=i+itemp;
      ftemp=fabs(freal(am[itemp1]))+fabs(fimag(am[itemp1]));
      if (ftemp > max) {
	max=ftemp;
	imax=i;
      }
    }
    if (max < eps) return (1);
    cmax=am[imax+itemp];
    if (imax != k) {
      for (j=k; j < n1; j++) {
	itemp1=k+j*n;
	itemp2=imax+j*n;
	ctemp=am[itemp1];
	am[itemp1]=am[itemp2];
	am[itemp2]=ctemp;
      }
    }
    for (j=k; j < n1; j++) {
      itemp1=k+j*n;
      am[itemp1]=fcmplx2op(am[itemp1],cmax,'/');
    }

    for (i=k+1; i < n; i++) {
      for (j=k+1; j < n1; j++) {
	itemp1=i+j*n;
	am[itemp1] = fcmplx2op(am[itemp1],fcmplx2op(am[k+j*n],am[i+itemp],'*'),'-');
      }
    }
  }
  
  itemp=n*n;
  x[n_1]=fcmplx2op(am[itemp+n_1],am[itemp-1],'/');
  for (k=n_1-1; k >= 0; k--) {
    cmax=am[k+itemp];
    for (j=k+1; j < n; j++) 
      cmax = fcmplx2op(cmax,fcmplx2op(x[j],am[k+j*n],'*'),'-');
    x[k]=cmax;
  }
 
  free((char *)am);
  return(0);
}

/* compute a*exp(j*phi) in a FFT fashion */    
int
fcvectexp(a,phi,n,npoints,ce)
float *a, *phi;
int n, npoints;
complex *ce;
{
  int i,k;
  float ftemp1,ftemp2,ftemp,dphi,dphi2;

  ftemp=TWO_PI/(float)npoints;
  fvectfunc(phi,n,fphase,phi);
  for (i=0; i < n; i++) {
    k=phi[i]/ftemp;
    dphi=phi[i]-ftemp*k;
    dphi2=dphi*dphi;
    k=k%npoints;
    ftemp1=a[i]*(1.0-0.5*dphi2);
    ftemp2=a[i]*(1.0-dphi2/6.0)*dphi;
    ce[i]=fcmplx2op(fcmplx(ftemp1,ftemp2),exp_table[k],'*');
  }

  return(0);
}

int
fcvectexp2(a,phi,n,npoints,ce)
float *a, *phi;
int n, npoints;
complex *ce;
{
  int i;

  for (i=0; i < n; i++) 
    ce[i]=fcmplx(a[i]*rcos(phi[i]),a[i]*rsin(phi[i]));

  return(0);
}

/* compute sinc(x)=sin(x)/x function in a FFT fashion */
int
fvectsinc(x,n,npoints,y)
float *x;
int n;
int npoints;
float *y;
{
  int i,k;
  float ftemp1,ftemp2,ftemp,dx,dx2;

  ftemp=TWO_PI/(float)npoints;
  for (i=0; i < n; i++) {
    k=x[i]/ftemp+0.5;
    dx=x[i]-ftemp*k;
    dx2=dx*dx;
    k=k%npoints;
    if (k < 0) k=npoints+k;
    if (k == 0) {
      ftemp1=0.0;
      ftemp2=1.0-dx2/6.0;
    }
    else {
      ftemp1=fimag(exp_table[k])*(1.0-0.5*dx2)/x[i];
      ftemp2=freal(exp_table[k])*(1.0-dx2/6.0)*dx/x[i];
    }
    y[i]=ftemp1+ftemp2;
  }

  return(0);
}

int 
fcvectFFTcorr(x,y,n,z)
complex *x,*y;
int n;
complex *z;
{
  complex *cvx,*cvy;

  cvx=(complex *)malloc((unsigned)(n*sizeof(complex)));
  cvy=(complex *)malloc((unsigned)(n*sizeof(complex)));
  if ((cvx == NULL) || (cvy == NULL)) {
    (void)fprintf(stderr,"Cannot allocate memory. \n");
    return (-1);
  }
  fcvectFFTforward2(x,n,cvx);
  fcvectFFTforward2(y,n,cvy);
  fcvect2op(cvx,cvy,n,'@',cvx);
  fcvectFFTinverse2(cvx,n,z);
  
  free((char *)cvx);
  free((char *)cvy);
  return (0);
}

int
fcvectSHIFT(x,n,nshift,y)
complex *x;
int n,nshift;
complex *y;
{
  int i,itemp;

  for (i=0; i < n; i++) {
    itemp=icycle(i,nshift,n);
    y[i]=x[itemp];
  }

  return (0);
}

int 
fft(n,y)
int n;
complex *y;
{
  int n1,logn,i,j,k,l1,l,logl,p;
  complex ctemp;
  
  logn=ilog2(n);
  n1=n >> 1;
  j=logn-1;
  k=0;
  for (logl=0; logl < logn; logl++) {
    do {
      for (i=0; i < n1; i++) {
	p=bitreversal((k >> j),logn);
	l=k+n1;
	ctemp=fcmplx2op(y[l],co_table[p],'@');
	y[l]=fcmplx2op(y[k],ctemp,'-');
	y[k]=fcmplx2op(ctemp,y[k],'+');
	k++;
      };
      k+=n1;
    } while (k < n);
    k=0;
    j--;
    n1=n1>>1;
  }

  for (i=1; i < n; i++) {
    k=bitreversal(i,logn);
    if (i > k) {
      ctemp=y[i];
      y[i]=y[k];
      y[k]=ctemp;
    }
  }
  return (0);
}

int 
ifft(n,y)
int n;
complex *y;
{
  int n1,logn,i,j,k,l,l1,logl,p;
  complex ctemp;
  
  logn=ilog2(n);
  n1=n >> 1;
  j=logn-1;
  k=0;
  for (logl=0; logl < logn; logl++) {
    do {
      for (i=0; i < n1; i++) {
	p=bitreversal((k >> j),logn);
	l=k+n1;
	ctemp=fcmplx2op(y[l],co_table[p],'*');
	y[l]=fcmplx2op(y[k],ctemp,'-');
	y[k]=fcmplx2op(ctemp,y[k],'+');
	k++;
      };
      k +=n1;
    } while (k < n);
    k=0;
    j--;
    n1=n1>>1;
  }

  for (i=1; i < n; i++) {
    k=bitreversal(i,logn);
    if (i > k) {
      ctemp=y[i];
      y[i]=y[k];
      y[k]=ctemp;
    }
  }
  
  for (i=0; i < n; i++)
    y[i]=fcmplx2op(y[i],fcmplx(1.0/n,0.0),'*');

  return (0);
}

